
"""

Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.

"""

import logging
import sys

from cliff.app import App
from cliff.commandmanager import CommandManager


class MusubiApp(App):

    log = logging.getLogger(__name__)

    def __init__(self):
        super(MusubiApp, self).__init__(
            description='Musubi DNSBL checker',
            version='0.2',
            command_manager=CommandManager('musubi.cli'),
        )

    def initialize_app(self, argv):
        self.log.debug('initialize_app')

    def prepare_to_run_command(self, cmd):
        self.log.debug('prepare_to_run_command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err):
        self.log.debug('clean_up %s', cmd.__class__.__name__)
        if err:
            self.log.debug('got an error: %s', err)


def main(argv=sys.argv[1:]):
    myapp = MusubiApp()
    if len(argv):
        try:
            return myapp.run(argv)
        except ValueError, err:
            #Command does not exist
            MusubiApp.log.error(err)
            sys.exit(2)
        except KeyboardInterrupt, err:
            MusubiApp.log.error(err)
            pass
    else:
        return myapp.run(['-h'])


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))





"""

Testing various DNS related utility functions

Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.

"""



from musubi.netdns import build_query

def test_build_query():
    '''Reverse the ip and append the name server
    reverse = '.'.join(reversed(ip.split('.')))
    query = '{reverse}.{dnsbl}.'.format(reverse=reverse, dnsbl=dnsbl)
    '''

    assert build_query('127.0.0.1', 'domain.com') == '1.0.0.127.domain.com.'
    assert build_query('127.0.0.1', 'domain.com.') != '1.0.0.127.domain.com.'
    assert build_query('1.1.1.1', 'domain.com') != '1.0.0.127.domain.com.'
    assert build_query('2.2.2.2', 'domain.com') == '2.2.2.2.domain.com.'
    assert (build_query('foobar', 'domain.com') != '2.2.2.2.domain.com.')
    





"""

Look up IP addresses for all mail servers in MX records

Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.

"""

import logging

from cliff.lister import Lister
from .netdns import get_mx_hosts, ips_from_domains


class GetIPs(Lister):

    """Show IP addresses for all servers in MX records for given domain"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Get parser options"""
        parser = super(GetIPs, self).get_parser(prog_name)
        parser.add_argument('ips', nargs='?', default=None)
        return parser

    def take_action(self, parsed_args):
        """Take action on args"""
        hosts = get_mx_hosts(parsed_args.ips)
        return (('IPs for MX Servers',),
               ((ip,) for ip in ips_from_domains(hosts)))




"""

Various DNS related utility functions

Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.

"""
import logging

import IPy
from dns import resolver, reversename

log = logging.getLogger(__name__)


def verify_domain(domain):
    '''Verify if host name exists'''
    try:
        hosts = get_mx_hosts(domain)
        if len(hosts):
            return True
    except resolver.NXDOMAIN:
        log.debug('NXDOMAIN: %s' % domain)
        return
    except resolver.NoAnswer:
        log.debug('NoAnswer: %s' % domain)
        return


def build_query(ip, dnsbl):
    '''Reverse the ip and append the name server'''
    reverse = '.'.join(reversed(ip.split('.')))
    query = '{reverse}.{dnsbl}.'.format(reverse=reverse, dnsbl=dnsbl)
    log.debug('Query: %s' % query)
    return query


def get_ips(domain):
    """Return all IPs (A records) for domain"""
    return [str(rdata) for rdata in resolver.query(domain, 'A')]


def get_mx_hosts(domain):
    """Get all domain names for servers in MX records"""
    return [x.exchange for x in resolver.query(domain, 'MX')]


def get_txt(domain):
    """Return all TXT records for a domain"""
    return [rdata.to_text() for rdata in resolver.query(domain, 'TXT')]


def get_spf(domain):
    """Return SPF TXT records for given domain"""
    txts = get_txt(domain)
    for txt in txts():
        test_txt = txt.lower()
        # TODO Need to look at SPF spec on this to check about spaces & quotes:
        if 'v=spf1' in test_txt[0:9]:
            yield txt


def ptr_from_ip(ip):
    """return PTR domain for given IP"""
    domain = reversename.from_address(ip)
    log.debug("domain ptr_from_ip: %s" % domain)
    return str(resolver.query(domain, "PTR")[0])


def ips_from_domains(domains):
    """
    Return unique set of all ips for all domains given.
    This is used to determine all the IPs we need to check
    for all mail servers we can find through MX records.

    Sometimes two or more servers have the same IP.

    $ musubi mx lancterprise.com

    +----------+-----------------------------+
    | priority | mail server domain name     |
    +----------+-----------------------------+
    | 10       | mailstore1.secureserver.net |
    | 0        | smtp.secureserver.net       |
    +----------+-----------------------------+

    $ musubi ips lancnterprise.com

    216.69.126.209

    """

    ips = []
    for host in domains:
        for ip in get_ips(host):
            if ip not in ips:
                ips.append(ip)
    return ips


def net_calc(iprange):
    """

    all_ips =IPy.IP('213.170.64.0/23')
    for ip in all_ips:
        print ip

    Note: We can get rid of IPy dependency when I figure out how to do
    this with pythondns.
    """
    return IPy.IP(iprange)




"""

Scan multiple DNSBLs for IP addresss or domain.

Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.


If you give the domain, musubi will try to find all your IP addresses
for each mail server by querying MX DNS records and then doing a lookup
for the IPs. If your mail server uses round-robin DNS, this of course
won't find all the IPs. You must find out the IP CIDR range and then
give that, e.g.

   musubi scan 192.0.64.0/24


"""

import sys
import logging

import dns
from cliff.lister import Lister
from IPy import IP
import requests

from .dnsbl import Base
from .netdns import get_mx_hosts, ips_from_domains, get_txt, build_query, \
    net_calc, verify_domain

requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)


DNSBL_LIST = 'http://musubi.cakebread.info/dnsbl.txt'

# Try to get list of working DNSBLs checked hourly, experimental.
# TODO Add options to use local list, pipe in, etc.
req = requests.get(DNSBL_LIST)
if req.status_code == 200:
    BASE_DNSBLS = req.text.split()
else:
    from .dnsbllist import BASE_DNSBLS


class Scan(Lister):

    """Scan multiple DNSBLs by IP or domain"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Scan, self).get_parser(prog_name)
        parser.add_argument('scan', nargs='?', default=None)
        return parser

    def dnsbl_check(self, ip):
        backend = Base(ip=ip, dnsbls=BASE_DNSBLS)
        return backend.check()

    def dnsbl_scanner(self, rdata, ip):
        for dnsbl, blacklisted in self.dnsbl_check(ip):
            # Scan.log.debug('Testing: %s' % dnsbl)
            if blacklisted:
                Scan.log.debug('blacklisted: %s' % dnsbl)
                try:
                    query = build_query(ip, dnsbl)
                    txt = get_txt(query)[0]
                except dns.resolver.NoAnswer:
                    Scan.log.debug("No TXT record for %s" % query)
                rdata.append(
                    (ip,
                     dnsbl,
                     blacklisted,
                     txt,)
                )
        return rdata

    def take_action(self, parsed_args):
        """This could be a lot prettier if I used these as arguments
        instead of trying to detect input type --IP --domain --range
        It's just easier to use without them, hmm.
        """
        arg = parsed_args.scan
        rdata = []
        if "/" in arg:
            # CIDR notation
            ips = net_calc(arg)
        else:
            try:
                # Throw exception if it's not an IP and then try domain name
                ip = IP(arg)
                ips = [ip]
            except ValueError:
                if verify_domain(arg):
                    hosts = get_mx_hosts(arg)
                    ips = ips_from_domains(hosts)
                else:
                    raise RuntimeError('Can not lookup domain: %s' % arg)
        for ip in ips:
            ip = str(ip)
            rdata = self.dnsbl_scanner(rdata, ip)

        if not len(rdata):
            Scan.log.debug("Not found on any DNSBL lists.")
            sys.exit(0)
        Scan.log.debug(rdata)
        return (('IP', 'DNSBL Host', 'Response Code', 'DNS TXT Record'), rdata)





BASE_DNSBLS = [
    'aspews.ext.sorbs.net',
    'b.barracudacentral.org',
    'bl.deadbeef.com',
    'bl.emailbasura.org',
    'bl.spamcannibal.org',
    'bl.spamcop.net',
    'blackholes.five-ten-sg.com',
    'blacklist.woody.ch',
    'bogons.cymru.com',
    'cbl.abuseat.org',
    'cdl.anti-spam.org.cn',
    'combined.abuse.ch',
    'combined.rbl.msrbl.net',
    'db.wpbl.info',
    'dnsbl-1.uceprotect.net',
    'dnsbl-2.uceprotect.net',
    'dnsbl-3.uceprotect.net',
    'dnsbl.ahbl.org',
    'dnsbl.inps.de',
    'dnsbl.njabl.org',
    'dnsbl.sorbs.net',
    'dnsbl.tornevall.org',
    'drone.abuse.ch',
    'duinv.aupads.org',
    'dul.dnsbl.sorbs.net',
    'dul.ru',
    'dyna.spamrats.com',
    'dynip.rothen.com',
    'http.dnsbl.sorbs.net'
    'images.rbl.msrbl.net',
    'ips.backscatterer.org',
    'ix.dnsbl.manitu.net',
    'korea.services.net',
    'misc.dnsbl.sorbs.net',
    'niku.2ch.net',
    'noptr.spamrats.com',
    'ohps.dnsbl.net.au',
    'omrs.dnsbl.net.au',
    'orvedb.aupads.org',
    'osps.dnsbl.net.au',
    'osrs.dnsbl.net.au',
    'owfs.dnsbl.net.au',
    'owps.dnsbl.net.au'
    'pbl.spamhaus.org',
    'phishing.rbl.msrbl.net',
    'probes.dnsbl.net.au'
    'proxy.bl.gweep.ca',
    'proxy.block.transip.nl',
    'psbl.surriel.com',
    'rbl.interserver.net',
    'rdts.dnsbl.net.au',
    'relays.bl.gweep.ca',
    'relays.bl.kundenserver.de',
    'relays.nether.net',
    'residential.block.transip.nl',
    'ricn.dnsbl.net.au',
    'rmst.dnsbl.net.au',
    'sbl.spamhaus.org',
    'short.rbl.jp',
    'smtp.dnsbl.sorbs.net',
    'socks.dnsbl.sorbs.net',
    'spam.abuse.ch',
    'spam.dnsbl.sorbs.net',
    'spam.rbl.msrbl.net',
    'spam.spamrats.com',
    'spamlist.or.kr',
    'spamrbl.imp.ch',
    't3direct.dnsbl.net.au',
    'tor.ahbl.org',
    'tor.dnsbl.sectoor.de',
    'torserver.tor.dnsbl.sectoor.de',
    'ubl.lashback.com',
    'ubl.unsubscore.com',
    'virbl.bit.nl',
    'virus.rbl.jp',
    'virus.rbl.msrbl.net',
    'web.dnsbl.sorbs.net',
    'wormrbl.imp.ch',
    'xbl.spamhaus.org',
    'zen.spamhaus.org',
    'zombie.dnsbl.sorbs.net',
]




"""
Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS
    OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Based on dnsbl.py by Vincent Carney (c) 2012

https://github.com/vincecarney/dnsbl/blob/master/dnsbl.py

Rewriting to do an additional TXT lookup of reason for blacklisting from DNSBL
using dnspython because gevent's socket can't lookup TXT records, apparently
it only does common types like A and CNAMEs etc.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""

import logging

import gevent
from gevent import socket


class Base(object):

    """DNSBL checker using gevent for super speed"""

    log = logging.getLogger(__name__)

    def __init__(self, ip=None, dnsbls=[], timeout=2):
        #TODO: Make timeout an --option
        self.ip = ip
        self.dnsbls = dnsbls
        self.timeout = timeout
        Base.log.debug('Checking IP %s' % ip)

    def build_query(self, dnsbl):
        '''Reverse the ip and append the name server'''
        reverse = '.'.join(reversed(self.ip.split('.')))
        return '{reverse}.{dnsbl}.'.format(reverse=reverse,
                                           dnsbl=dnsbl)

    def query(self, dnsbl):
        '''Perform query using gevent'''
        try:
            result = socket.gethostbyname(self.build_query(dnsbl))
        except socket.gaierror as err:
            result = False
            err = str(err)
            if not 'Errno 3' in err:
                Base.log.debug('Base.query.exception: %s %s' % (dnsbl, err))
        return dnsbl, result

    def check(self):
        '''Check results of all gevent jobs'''
        results = []
        jobs = [gevent.spawn(self.query, dnsbl)
                for dnsbl in self.dnsbls]
        gevent.joinall(jobs, self.timeout)
        for job in jobs:
            if job.successful():
                results.append(job.value)
            else:
                results.append((job.args[0], None))
        return results




"""

Look up SPF records for a domain

Copyright (c) 2012, 2013 Rob Cakebread
All rights reserved.

"""

import logging

from cliff.command import Command
import dns.resolver


class GetSPF(Command):

    """Show SPF records for domain"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(GetSPF, self).get_parser(prog_name)
        parser.add_argument('getspf', nargs='?', default=None)
        return parser

    def take_action(self, parsed_args):
        domain = parsed_args.getspf
        for x in dns.resolver.query(domain, 'TXT'):
            txt = x.to_text()
            test_txt = txt.lower()
            # There are many uses for TXT, we only want SPF records:
            if 'v=spf1' in test_txt[0:7]:
                print(txt)



#!/usr/bin/env python

PROJECT = 'musubi'
VERSION = '0.2'

import distribute_setup
distribute_setup.use_setuptools()
from setuptools import setup, find_packages


try:
    long_description = open('README.rst', 'rt').read()
except IOError:
    long_description = 'Uh oh, we may need a new hard drive.'

setup(
    name=PROJECT,
    version=VERSION,
    description='Musubi is a command-line DNSBL checker and MX toolkit.',
    long_description=long_description,
    author='Rob Cakebread',
    author_email='cakebread@gmail.com',
    url='https://github.com/cakebread/musubi',
    download_url='https://github.com/cakebread/musubi/tarball/master',
    classifiers=['Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.7',
                 'Intended Audience :: Developers',
                 'Environment :: Console',
                 ],
    platforms=['Any'],
    scripts=[],
    provides=[],
    install_requires=['requests', 'dnspython', 'IPy', 'distribute',
        'cliff', 'cliff-tablib', 'gevent', 'greenlet'],
    namespace_packages=[],
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'musubi = musubi.main:main'
        ],
        'musubi.cli': [
            'ips = musubi.ips:GetIPs',
            'mx = musubi.mx:GetMX',
            'spf = musubi.spf:GetSPF',
            'scan = musubi.scan:Scan',
        ],
    },

    zip_safe=False,
)




"""

Look up MX records for a domain, show IP addresses, PTRs

Copyright (c) 2012, 2013, Rob Cakebread
All rights reserved.

"""

import logging
import socket

from cliff.lister import Lister
import dns.resolver

from .netdns import verify_domain


class GetMX(Lister):

    """Show MX records for domain"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Set up parser options for this command"""
        parser = super(GetMX, self).get_parser(prog_name)
        parser.add_argument('getmx', nargs='?', default=None)
        return parser

    def take_action(self, parsed_args):
        """Take action for this command"""
        domain = parsed_args.getmx
        if verify_domain(domain):
            return (('Priority', 'Mail Server Domain Name', 'IP', 'PTR'),
                   ((
                    x.preference,
                    str(x.exchange).lower(),
                    dns.resolver.query(str(x.exchange), 'A')[0],
                    socket.gethostbyaddr(
                    str(dns.resolver.query(
                        str(x.exchange), 'A')[0]))[0]
                    ) for x in dns.resolver.query(domain, 'MX')))
        else:
            raise RuntimeError('Can not lookup domain: %s' % domain)
